@@ -30,7 +30,6 @@
 #include <string.h>
 #include <camel/camel-i18n.h>
 #include <libedataserverui/e-passwords.h>
-#include "e-util/e-error.h"
 #include <libedataserver/e-account.h>
 #include <libecal/e-cal.h>
 #include <libedataserver/e-account-list.h>
@@ -50,6 +49,8 @@ LIMBAPI_CFLAGS or something is going wrong */
 
 #define d(x) x
 
+G_DEFINE_TYPE (ExchangeMAPIAccountListener, exchange_mapi_account_listener, G_TYPE_OBJECT)
+
 struct _ExchangeMAPIAccountListenerPrivate {
 	GConfClient *gconf_client;
 	/* we get notification about mail account changes from this object */
@@ -69,8 +70,6 @@ struct _ExchangeMAPIAccountInfo {
 /* list of ExchangeMAPIAccountInfo structures */
 static 	GList *mapi_accounts = NULL;
 
-#define PARENT_TYPE G_TYPE_OBJECT
-
 static GObjectClass *parent_class = NULL;
 
 static void 
@@ -114,7 +113,7 @@ exchange_mapi_account_listener_class_init (ExchangeMAPIAccountListenerClass *cla
 {
 	GObjectClass *object_class;
 	
-	parent_class = g_type_class_ref (PARENT_TYPE);
+	parent_class = g_type_class_ref (G_TYPE_OBJECT);
 	object_class = G_OBJECT_CLASS (class);
 	
 	/* virtual method override */
@@ -123,7 +122,7 @@ exchange_mapi_account_listener_class_init (ExchangeMAPIAccountListenerClass *cla
 }
 
 static void 
-exchange_mapi_account_listener_init (ExchangeMAPIAccountListener *config_listener, ExchangeMAPIAccountListenerClass *class)
+exchange_mapi_account_listener_init (ExchangeMAPIAccountListener *config_listener)
 {
 	config_listener->priv = g_new0 (ExchangeMAPIAccountListenerPrivate, 1);
 }
@@ -452,6 +451,13 @@ add_addressbook_sources (EAccount *account, GSList *folders)
 		e_source_set_property (source, "offline_sync", 
 					       camel_url_get_param (url, "offline_sync") ? "1" : "0");
 		e_source_set_property (source, "completion", "true");
+		if (folder->is_default) 
+			e_source_set_property (source, "delete", "no");
+		if (folder->parent_folder_id) {
+			gchar *tmp = exchange_mapi_util_mapi_id_to_string (folder->parent_folder_id);
+			e_source_set_property (source, "parent-fid", tmp);
+			g_free (tmp);
+		}
 		e_source_group_add_source (group, source, -1);
 		g_object_unref (source);
 	}
@@ -477,6 +483,7 @@ add_addressbook_sources (EAccount *account, GSList *folders)
 		// camel_url_get_param (url, "offline_sync") ? "1" : "0");
 		e_source_set_property(source, "offline_sync", "1");
 		e_source_set_property (source, "completion", "true");
+		e_source_set_property (source, "delete", "no");
 		e_source_group_add_source (group, source, -1);
 		g_object_unref (source);
 	}
@@ -600,7 +607,7 @@ mapi_account_removed (EAccountList *account_listener, EAccount *account)
 }
 
 static gboolean
-create_profile_entry (CamelURL *url)
+create_profile_entry (CamelURL *url, EAccount *account)
 {
 	gboolean status = FALSE;
 	guint8 attempts = 0; 
@@ -611,7 +618,7 @@ create_profile_entry (CamelURL *url)
 		key = camel_url_to_string (url, CAMEL_URL_HIDE_PASSWORD | CAMEL_URL_HIDE_PARAMS);
 		password = e_passwords_get_password (EXCHANGE_MAPI_PASSWORD_COMPONENT, key);
 		if (!password) {
-			gboolean remember = FALSE;
+			gboolean remember = account && e_account_get_bool (account, E_ACCOUNT_SOURCE_SAVE_PASSWD);
 			gchar *title;
 
 			title = g_strdup_printf (_("Enter Password for %s@%s"), url->user, url->host);
@@ -674,7 +681,7 @@ mapi_account_changed (EAccountList *account_listener, EAccount *account)
 
 	if (existing_account_info == NULL && isa_mapi_account) {
 		/* some account of other type is changed to MAPI */
-		if (create_profile_entry (new_url)) {
+		if (create_profile_entry (new_url, account)) {
 			/* Things are successful */
 			gchar *profname = NULL, *uri = NULL; 
 			ExchangeMAPIAccountListener *config_listener = exchange_mapi_accounts_peek_config_listener();
@@ -707,7 +714,7 @@ mapi_account_changed (EAccountList *account_listener, EAccount *account)
 		} else if (!mapi_camel_url_equal (old_url, new_url) || (existing_account_info->enabled != account->enabled)) {
 		/* Some or all of the account info changed OR the account has been moved from a disabled state to enabled state */
 			mapi_account_removed (account_listener, account);
-			if (create_profile_entry (new_url)) {
+			if (create_profile_entry (new_url, account)) {
 				/* Things are successful */
 				gchar *profname = NULL, *uri = NULL; 
 				ExchangeMAPIAccountListener *config_listener = exchange_mapi_accounts_peek_config_listener();
@@ -763,30 +770,8 @@ exchange_mapi_account_listener_construct (ExchangeMAPIAccountListener *config_li
 	g_signal_connect (config_listener->priv->account_list, "account_removed", G_CALLBACK (mapi_account_removed), NULL);
 }
 
-GType
-exchange_mapi_account_listener_get_type (void)
-{
-	static GType exchange_mapi_account_listener_type = 0;
-
-	if (!exchange_mapi_account_listener_type) {
-		static GTypeInfo info = {
-			sizeof (ExchangeMAPIAccountListenerClass),
-			(GBaseInitFunc) NULL,
-			(GBaseFinalizeFunc) NULL,
-			(GClassInitFunc) exchange_mapi_account_listener_class_init,
-			NULL, NULL,
-			sizeof (ExchangeMAPIAccountListener),
-			0,
-			(GInstanceInitFunc) exchange_mapi_account_listener_init
-		};
-		exchange_mapi_account_listener_type = g_type_register_static (PARENT_TYPE, "ExchangeMAPIAccountListener", &info, 0);
-	}
-
-	return exchange_mapi_account_listener_type;
-}
-
 ExchangeMAPIAccountListener *
-exchange_mapi_account_listener_new ()
+exchange_mapi_account_listener_new (void)
 {
 	ExchangeMAPIAccountListener *config_listener;
 
