@@ -79,12 +79,6 @@ typedef struct {
 } flags_diff_t;
 
 /*For collecting summary info from server*/
-typedef struct {
-	GSList *items_list;
-	GTimeVal last_modification_time;
-	CamelFolder *folder;
-	CamelFolderChangeInfo *changes;
-}fetch_items_data;
 
 static CamelMimeMessage *mapi_folder_item_to_msg( CamelFolder *folder, MapiItem *item, CamelException *ex );
 static void mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **changeinfo,
@@ -104,6 +98,22 @@ mapi_folder_search_by_expression (CamelFolder *folder, const char *expression, C
 	return matches;
 }
 
+static GPtrArray *
+mapi_folder_search_by_uids (CamelFolder *folder, const gchar *expression, GPtrArray *uids, CamelException *ex)
+{
+	GPtrArray *matches;
+	CamelMapiFolder *mapi_folder = CAMEL_MAPI_FOLDER (folder);
+
+	if (uids->len == 0)
+		return g_ptr_array_new ();
+
+	CAMEL_MAPI_FOLDER_LOCK (mapi_folder, search_lock);
+	camel_folder_search_set_folder (mapi_folder->search, folder);
+	matches = camel_folder_search_search (mapi_folder->search, expression, uids, ex);
+	CAMEL_MAPI_FOLDER_UNLOCK (mapi_folder, search_lock);
+
+	return matches;
+}
 
 static int
 mapi_getv (CamelObject *object, CamelException *ex, CamelArgGetV *args)
@@ -178,10 +188,8 @@ mapi_refresh_info(CamelFolder *folder, CamelException *ex)
 
 }
 
-
-/*Using GFunc*/
-static void 
-mapi_item_free (MapiItem *item, gpointer data)
+void 
+mapi_item_free (MapiItem *item)
 {
 	g_free (item->header.subject);
 	g_free (item->header.from);
@@ -196,6 +204,9 @@ mapi_item_free (MapiItem *item, gpointer data)
 
 	exchange_mapi_util_free_attachment_list (&item->attachments);
 	exchange_mapi_util_free_stream_list (&item->generic_streams);
+	exchange_mapi_util_free_recipient_list (&item->recipients);
+
+	g_free (item);
 }
 
 static gboolean
@@ -283,7 +294,7 @@ read_item_common (MapiItem *item, uint32_t ulPropTag, gconstpointer prop_data)
 }
 
 static gboolean
-fetch_items_cb (FetchItemsCallbackData *item_data, gpointer data)
+fetch_items_summary_cb (FetchItemsCallbackData *item_data, gpointer data)
 {
 	fetch_items_data *fi_data = (fetch_items_data *)data;
 	
@@ -351,8 +362,8 @@ fetch_items_cb (FetchItemsCallbackData *item_data, gpointer data)
 	fi_data_mod_time.tv_usec = fi_data->last_modification_time.tv_usec;
 
 	if (timeval_compare (&item_modification_time, &fi_data_mod_time) == 1) {
-			fi_data->last_modification_time.tv_sec = item_modification_time.tv_sec;
-			fi_data->last_modification_time.tv_usec = item_modification_time.tv_usec;
+		fi_data->last_modification_time.tv_sec = item_modification_time.tv_sec;
+		fi_data->last_modification_time.tv_usec = item_modification_time.tv_usec;
 	}
 
 	if ((*flags & MSGFLAG_READ) != 0)
@@ -363,7 +374,8 @@ fetch_items_cb (FetchItemsCallbackData *item_data, gpointer data)
 	*slist = g_slist_prepend (*slist, item);
 
 	/*Write summary to db in batches of SUMMARY_FETCH_BATCH_COUNT items.*/ 
-	if ( (item_data->index % SUMMARY_FETCH_BATCH_COUNT == 0) || item_data->index == item_data->total-1) {
+	if ((item_data->index % SUMMARY_FETCH_BATCH_COUNT == 0) ||
+	     item_data->index == item_data->total-1) {
 		mapi_update_cache (fi_data->folder, *slist, &fi_data->changes, NULL, false);
 		g_slist_foreach (*slist, (GFunc)mapi_item_free, NULL);
 		g_slist_free (*slist);
@@ -483,7 +495,7 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
 	for ( ; item_list != NULL ; item_list = g_slist_next (item_list) ) {
 		MapiItem *temp_item ;
 		MapiItem *item;
-		gchar *msg_uid, *to = NULL, *from = NULL, *cc = NULL;
+		gchar *msg_uid;
 		guint64 id;
 
 		exists = FALSE;
@@ -522,6 +534,7 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
 		if (!exists) {
  			GSList *l = NULL;
  			guint32 count_to = 0, count_cc =0;
+			gchar *to = NULL, *cc = NULL;
 
 			mi->info.uid = exchange_mapi_util_mapi_ids_to_uid(item->fid, item->mid);
 			mi->info.subject = camel_pstring_strdup(item->header.subject);
@@ -563,8 +576,10 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
 					switch (*type) {
 					case MAPI_TO:
 						if (count_to) {
+							gchar *tmp = to;
 							to = g_strconcat (to, ", ", formatted_id, NULL);
 							g_free (formatted_id);
+							g_free (tmp);
 						} else
 							to = formatted_id;
 						count_to ++;
@@ -572,8 +587,10 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
 
 					case MAPI_CC:
 						if (count_cc) {
+							gchar *tmp = cc;
 							cc = g_strconcat (cc, ", ", formatted_id, NULL);
 							g_free (formatted_id);
+							g_free (tmp);
 						} else
 							cc = formatted_id;
 						count_cc ++;
@@ -598,15 +615,20 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
  				item->header.from_email : item->header.from;
 
 			if (item->header.from_email) {
-				from = camel_internet_address_format_address (item->header.from, 
- 									      item->header.from_email);
+				gchar *from = camel_internet_address_format_address (item->header.from, 
+										     item->header.from_email);
  				mi->info.from = camel_pstring_strdup (from);
+
+				g_free (from);
 			} else
 				mi->info.from = NULL;
 
 			/* Fallback */
- 			mi->info.to = to ? camel_pstring_strdup (to) : g_strdup (item->header.to);
- 			mi->info.cc = to ? camel_pstring_strdup (cc) : g_strdup (item->header.cc);
+ 			mi->info.to = to ? camel_pstring_strdup (to) : camel_pstring_strdup (item->header.to);
+ 			mi->info.cc = cc ? camel_pstring_strdup (cc) : camel_pstring_strdup (item->header.cc);
+
+			g_free (to);
+			g_free (cc);
 		}
 
 		if (exists) {
@@ -654,7 +676,6 @@ struct mapi_update_deleted_msg {
 
 	CamelFolder *folder;
 	mapi_id_t folder_id;
-	gboolean need_refresh;
 };
 
 static gboolean
@@ -728,8 +749,8 @@ mapi_sync_deleted (CamelSession *session, CamelSessionThreadMsg *msg)
 	for (index = 0; index < count; index++) {
 		GSList *tmp_list_item = NULL;
 
-		/* Iterate in a reverse order, thus removal will not hurt */
-		info = camel_folder_summary_index (m->folder->summary, count - index - 1);
+		/*FIXME :Any other list available ???*/
+		info = camel_folder_summary_index (m->folder->summary, index);
 		if (!info) continue; /*This is bad. *Should* not happen*/
 
 		uid = camel_message_info_uid (info);
@@ -759,11 +780,9 @@ mapi_sync_deleted (CamelSession *session, CamelSessionThreadMsg *msg)
 
 	camel_operation_end (NULL);
 
-	camel_object_trigger_event (m->folder, "folder_changed", changes);
+	/* camel_object_trigger_event (m->folder, "folder_changed", changes); */
 	camel_folder_change_info_free (changes);
 
-	m->need_refresh = camel_folder_summary_count (m->folder->summary) != g_slist_length (server_uid_list);
-
 	/* Discard server uid list */
 	g_slist_foreach (server_uid_list, (GFunc) g_free, NULL);
 	g_slist_free (server_uid_list);
@@ -774,25 +793,8 @@ mapi_sync_deleted_free (CamelSession *session, CamelSessionThreadMsg *msg)
 {
 	struct mapi_update_deleted_msg *m = (struct mapi_update_deleted_msg *)msg;
 
-	if (m->need_refresh) {
-		CamelMapiSummary *mapi_summary = CAMEL_MAPI_SUMMARY (m->folder->summary);
-		if (mapi_summary) {
-			CamelException ex = CAMEL_EXCEPTION_INITIALISER;
-
-			CAMEL_SERVICE_REC_LOCK (m->folder->parent_store, connect_lock);
-			g_free (mapi_summary->sync_time_stamp);
-			mapi_summary->sync_time_stamp = NULL;
-			CAMEL_SERVICE_REC_UNLOCK (m->folder->parent_store, connect_lock);
-
-			mapi_refresh_folder (m->folder, &ex);
-
-			if (camel_exception_is_set (&ex))
-				g_warning ("%s: %s", G_STRFUNC, ex.desc);
-			camel_exception_clear (&ex);
-		}
-	}
-
 	camel_object_unref (m->folder);
+	/* camel_session_thread_msg_free (session, &m->msg); */
 }
 
 static CamelSessionThreadOps deleted_items_sync_ops = {
@@ -808,7 +810,7 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 	CamelMessageInfo *info = NULL;
 	CamelMapiMessageInfo *mapi_info = NULL;
 
-	GSList *read_items = NULL, *unread_items = NULL;
+	GSList *read_items = NULL, *unread_items = NULL, *to_free = NULL;
 	flags_diff_t diff, unset_flags;
 	const char *folder_id;
 	mapi_id_t fid, deleted_items_fid;
@@ -849,40 +851,55 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 			mapi_id_t *mid = g_new0 (mapi_id_t, 1); /* FIXME : */
 			mapi_id_t temp_fid;
 			guint32 flags;
+			gboolean used = FALSE;
 
 			uid = camel_message_info_uid (info);
 			flags= camel_message_info_flags (info);
 
 			/* Why are we getting so much noise here :-/ */
-			if (!exchange_mapi_util_mapi_ids_from_uid (uid, &temp_fid, mid))
+			if (!exchange_mapi_util_mapi_ids_from_uid (uid, &temp_fid, mid)) {
+				g_free (mid);
 				continue;
+			}
 
 			mapi_utils_do_flags_diff (&diff, mapi_info->server_flags, mapi_info->info.flags);
 			mapi_utils_do_flags_diff (&unset_flags, flags, mapi_info->server_flags);
 
 			diff.changed &= folder->permanent_flags;
 			if (!diff.changed) {
-				camel_message_info_free(info);
+				camel_message_info_free (info);
+				g_free (mid);
 				continue;
 			} else {
 				if (diff.bits & CAMEL_MESSAGE_DELETED) {
-					if (diff.bits & CAMEL_MESSAGE_SEEN) 
+					if (diff.bits & CAMEL_MESSAGE_SEEN) {
 						read_items = g_slist_prepend (read_items, mid);
-					if (deleted_items)
+						used = TRUE;
+					}
+					if (deleted_items) {
 						deleted_items = g_slist_prepend (deleted_items, mid);
-					else {
+						used = TRUE;
+					} else {
 						g_slist_free (deleted_head);
 						deleted_head = NULL;
 						deleted_head = deleted_items = g_slist_prepend (deleted_items, mid);
+						used = TRUE;
 					}
 				}
 			}
 
 			if (diff.bits & CAMEL_MESSAGE_SEEN) {
 				read_items = g_slist_prepend (read_items, mid);
+				used = TRUE;
 			} else if (unset_flags.bits & CAMEL_MESSAGE_SEEN) {
 				unread_items = g_slist_prepend (unread_items, mid);
+				used = TRUE;
 			}
+
+			if (used)
+				to_free = g_slist_prepend (to_free, mid);
+			else
+				g_free (mid);
 		}
 		camel_message_info_free (info);
 	}
@@ -927,10 +944,11 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 		deleted_items = g_slist_next (deleted_items);
 	}
 
-	if (unread_items) {
-		/* TODO */
-		g_slist_free (unread_items);
-	}
+	g_slist_free (unread_items);
+	g_slist_free (deleted_head);
+
+	g_slist_foreach (to_free, (GFunc) g_free, NULL);
+	g_slist_free (to_free);
 
 	if (expunge) {
 		/* TODO */
@@ -941,26 +959,12 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 	CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
 }
 
-
-void
-mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
+gboolean
+camel_mapi_folder_fetch_summary (CamelStore *store, const mapi_id_t fid, struct mapi_SRestriction *res,
+				 struct SSortOrderSet *sort, fetch_items_data *fetch_data, guint32 options)
 {
-
-	CamelMapiStore *mapi_store = CAMEL_MAPI_STORE (folder->parent_store);
-	CamelMapiFolder *mapi_folder = CAMEL_MAPI_FOLDER (folder);
-	CamelMapiSummary *mapi_summary = CAMEL_MAPI_SUMMARY (folder->summary);
-	CamelSession *session = ((CamelService *)folder->parent_store)->session;
-
-	gboolean is_proxy = folder->parent_store->flags & CAMEL_STORE_PROXY;
-	gboolean is_locked = FALSE;
 	gboolean status;
-
-	struct mapi_SRestriction *res = NULL;
-	struct SSortOrderSet *sort = NULL;
-	struct mapi_update_deleted_msg *deleted_items_op_msg;
-	fetch_items_data *fetch_data = g_new0 (fetch_items_data, 1);
-
-	const gchar *folder_id = NULL;
+	CamelMapiStore *mapi_store = (CamelMapiStore *) store;
 
 	const guint32 summary_prop_list[] = {
 		PR_INTERNET_CPID,
@@ -989,6 +993,44 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 		PR_TRANSPORT_MESSAGE_HEADERS_UNICODE
 	};
 
+	/*TODO : Check for online state*/
+
+	camel_operation_start (NULL, _("Fetching summary information for new messages in")); /* %s"), folder->name); */
+
+	CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
+
+	status = exchange_mapi_connection_fetch_items  (fid, res, sort, summary_prop_list,
+							G_N_ELEMENTS (summary_prop_list), 
+							NULL, NULL, fetch_items_summary_cb,
+							fetch_data, options);
+
+	CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
+
+	camel_operation_end (NULL);
+
+	return status;
+}
+
+void
+mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
+{
+
+	CamelMapiStore *mapi_store = CAMEL_MAPI_STORE (folder->parent_store);
+	CamelMapiFolder *mapi_folder = CAMEL_MAPI_FOLDER (folder);
+	CamelMapiSummary *mapi_summary = CAMEL_MAPI_SUMMARY (folder->summary);
+	CamelSession *session = ((CamelService *)folder->parent_store)->session;
+
+	gboolean is_proxy = folder->parent_store->flags & CAMEL_STORE_PROXY;
+	gboolean is_locked = FALSE;
+	gboolean status;
+
+	struct mapi_SRestriction *res = NULL;
+	struct SSortOrderSet *sort = NULL;
+	struct mapi_update_deleted_msg *deleted_items_op_msg;
+	fetch_items_data *fetch_data = g_new0 (fetch_items_data, 1);
+
+	const gchar *folder_id = NULL;
+
 	if (((CamelOfflineStore *) mapi_store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL)
 		goto end1;
 
@@ -1064,14 +1106,8 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 		if (((CamelMapiFolder *)folder)->type & CAMEL_MAPI_FOLDER_PUBLIC)
 			options |= MAPI_OPTIONS_USE_PFSTORE;
 
-		camel_operation_start (NULL, _("Fetching summary information for new messages in %s"), folder->name);
-
-		status = exchange_mapi_connection_fetch_items  (temp_folder_id, res, sort,
-								summary_prop_list, G_N_ELEMENTS (summary_prop_list), 
-								NULL, NULL, 
-								fetch_items_cb, fetch_data, 
-								options);
-		camel_operation_end (NULL);
+		status = camel_mapi_folder_fetch_summary ((CamelStore *)mapi_store, temp_folder_id, res, sort,
+							  fetch_data, options);
 
 		if (!status) {
 			camel_exception_set (ex, CAMEL_EXCEPTION_SERVICE_INVALID, _("Fetching items failed"));
@@ -1079,6 +1115,7 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 		}
 
 		/*Preserve last_modification_time from this fetch for later use with restrictions.*/
+		g_free (mapi_summary->sync_time_stamp);
 		mapi_summary->sync_time_stamp = g_time_val_to_iso8601 (&fetch_data->last_modification_time);
 
 		camel_folder_summary_touch (folder->summary);
@@ -1092,7 +1129,6 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 							 sizeof (*deleted_items_op_msg));
 		deleted_items_op_msg->folder = folder;
 		deleted_items_op_msg->folder_id = temp_folder_id;
-		deleted_items_op_msg->need_refresh = FALSE;
 		camel_object_ref (folder);
 
 		camel_session_thread_queue (session, &deleted_items_op_msg->msg, 0);
@@ -1115,6 +1151,7 @@ static const uint32_t camel_GetPropsList[] = {
 	PR_MID, 
 	PR_INTERNET_CPID,
 
+
 	PR_TRANSPORT_MESSAGE_HEADERS,
 	PR_TRANSPORT_MESSAGE_HEADERS_UNICODE,
 	PR_MESSAGE_CLASS, 
@@ -1305,30 +1342,36 @@ fetch_item_cb (FetchItemsCallbackData *item_data, gpointer data)
 		}
 	}
 
+	item->is_cal = FALSE;
 	if (g_str_has_prefix (msg_class, IPM_SCHEDULE_MEETING_PREFIX)) {
 		guint8 *appointment_body_str = (guint8 *) exchange_mapi_cal_util_camel_helper (item_data->properties, 
 											     item_data->streams, 
 											     item_data->recipients, item_data->attachments);
 
-		body = g_new0(ExchangeMAPIStream, 1);
-		body->proptag = PR_BODY;
-		body->value = g_byte_array_new ();
-		body->value = g_byte_array_append (body->value, appointment_body_str, strlen ((const char *)appointment_body_str));
+		if (appointment_body_str && *appointment_body_str) {
+			body = g_new0(ExchangeMAPIStream, 1);
+			body->proptag = PR_BODY;
+			body->value = g_byte_array_new ();
+			body->value = g_byte_array_append (body->value, appointment_body_str, strlen ((const char *)appointment_body_str));
 
-		item->msg.body_parts = g_slist_append (item->msg.body_parts, body);
-
-		item->is_cal = TRUE;
+			item->msg.body_parts = g_slist_append (item->msg.body_parts, body);
+			item->is_cal = TRUE;
+		}
 
 		g_free (appointment_body_str);
-	} else { 
+	}
+
+	if (!item->is_cal) {
 		/* always prefer unicode version, as that can be properly read */
-		if (!((body = exchange_mapi_util_find_stream (item_data->streams, PR_BODY_UNICODE)) || 
-		      (body = exchange_mapi_util_find_stream (item_data->streams, PR_BODY))))
-			body = exchange_mapi_util_find_stream (item_data->streams, PR_HTML);
+		if (!(body = exchange_mapi_util_find_stream (item_data->streams, PR_BODY_UNICODE)))
+			body = exchange_mapi_util_find_stream (item_data->streams, PR_BODY);
 
-		item->msg.body_parts = g_slist_append (item->msg.body_parts, body);
+		if (body)
+			item->msg.body_parts = g_slist_append (item->msg.body_parts, body);
 
-		item->is_cal = FALSE;
+		body = exchange_mapi_util_find_stream (item_data->streams, PR_HTML);
+		if (body)
+			item->msg.body_parts = g_slist_append (item->msg.body_parts, body);
 	}
 
 	if (delivery_date) {
@@ -1352,7 +1395,7 @@ fetch_item_cb (FetchItemsCallbackData *item_data, gpointer data)
 
 
 static void
-mapi_msg_set_recipient_list (CamelMimeMessage *msg, MapiItem *item)
+mapi_mime_set_recipient_list (CamelMimeMessage *msg, MapiItem *item)
 {
 	GSList *l = NULL;
 	CamelInternetAddress *to_addr, *cc_addr, *bcc_addr;
@@ -1401,20 +1444,25 @@ mapi_msg_set_recipient_list (CamelMimeMessage *msg, MapiItem *item)
 			camel_internet_address_add (bcc_addr, display_name, recip->email_id);
 			break;
 		}
-		/* g_free (display_name); */
+
+		g_free (display_name);
 	}
-	
+
 	/*Add to message*/
-	/*Note : To field is added from PR_TRANSPORT_MESSAGE_HEADERS*/
-	camel_mime_message_set_recipients(msg, "Cc", cc_addr);
-	camel_mime_message_set_recipients(msg, "Bcc", bcc_addr);
+	/*Note : To field is added from PR_TRANSPORT_MESSAGE_HEADERS
+	  But, in sent_items folder we don't get TRANSPORT_MESSAGE_HEADERS */
+	if (!item->header.transport_headers) {
+		camel_mime_message_set_recipients(msg, "To", to_addr);
+		camel_mime_message_set_recipients(msg, "Cc", cc_addr);
+		camel_mime_message_set_recipients(msg, "Bcc", bcc_addr);
+	}
 
 	/*TODO : Unref *_addr ? */
 }
 
 
 static void
-mapi_populate_details_from_item (CamelFolder *folder, CamelMimeMessage *msg, MapiItem *item)
+mapi_mime_set_msg_headers (CamelFolder *folder, CamelMimeMessage *msg, MapiItem *item)
 {
 	char *temp_str = NULL;
 	const char *from_email;
@@ -1490,15 +1538,27 @@ mapi_populate_details_from_item (CamelFolder *folder, CamelMimeMessage *msg, Map
 		camel_internet_address_add(addr, item->header.from, item->header.from_email);
 		camel_mime_message_set_from(msg, addr);
 	}
+
+	/* Threading */
+	if (item->header.message_id)
+		camel_medium_add_header (CAMEL_MEDIUM (msg), "Message-ID", item->header.message_id);
+
+	if (item->header.references)
+		camel_medium_add_header (CAMEL_MEDIUM (msg), "References", item->header.references);
+
+	if (item->header.in_reply_to)
+		camel_medium_add_header (CAMEL_MEDIUM (msg), "In-Reply-To", item->header.in_reply_to);
+
 }
 
 
-static void
-mapi_populate_msg_body_from_item (CamelMimePart *part, MapiItem *item, ExchangeMAPIStream *body)
+static CamelMimePart *
+mapi_mime_msg_body (MapiItem *item, const ExchangeMAPIStream *body)
 {
+	CamelMimePart *part = camel_mime_part_new ();
 	camel_mime_part_set_encoding (part, CAMEL_TRANSFER_ENCODING_8BIT);
 	
-	if (body) {
+	if (body && body->value && body->value->len > 0) {
 		const gchar* type = NULL;
 		gchar *buff = NULL;
 
@@ -1520,9 +1580,14 @@ mapi_populate_msg_body_from_item (CamelMimePart *part, MapiItem *item, ExchangeM
 
 		g_free (buff);
 	} else
-		camel_mime_part_set_content (part, "", strlen (""), "text/plain");
+		camel_mime_part_set_content (part, " ", strlen (" "), "text/plain");
+
+	return part;
 }
 
+#if 0
+
+/* GCompareFunc. Used for ordering body types in a GSList.*/
 static gint
 sort_bodies_cb (gconstpointer a, gconstpointer b)
 {
@@ -1550,143 +1615,192 @@ sort_bodies_cb (gconstpointer a, gconstpointer b)
 	return aidx - bidx;
 }
 
+#endif
+
+/* Adds parts to multipart. Convenience function. */
+static void
+mapi_mime_multipart_add_attachments (CamelMultipart *multipart, GSList *attachs)
+{
+	CamelMimePart *part;
+	while (attachs) {
+		part = attachs->data;
+		camel_multipart_add_part (multipart, part);
+		camel_object_unref (part);
+		attachs = attachs->next;
+	}
+}
+
+/* Process body stream and related objects into a MIME mulitpart */
+static CamelMultipart *
+mapi_mime_build_multipart_related (MapiItem *item, const ExchangeMAPIStream *stream,
+				   GSList *inline_attachs)
+{
+	CamelMimePart *part;
+	CamelMultipart *m_related = camel_multipart_new ();
+	camel_data_wrapper_set_mime_type (CAMEL_DATA_WRAPPER (m_related), "multipart/related");
+	camel_multipart_set_boundary (m_related, NULL);
+	
+	part = mapi_mime_msg_body (item, stream);
+	camel_multipart_add_part (m_related, part);
+	camel_object_unref (part);
+
+	mapi_mime_multipart_add_attachments (m_related, inline_attachs);
+
+	return m_related;
+}
+
+/* Process multiple body types and pack them in a MIME mulitpart */
+static CamelMultipart *
+mapi_mime_build_multipart_alternative (MapiItem *item, GSList *body_parts, GSList *inline_attachs)
+{
+	CamelMimePart *part;
+	CamelMultipart *m_alternative = camel_multipart_new ();
+	camel_data_wrapper_set_mime_type (CAMEL_DATA_WRAPPER (m_alternative), 
+					  "multipart/alternative");
+	camel_multipart_set_boundary (m_alternative, NULL);
+
+	while (body_parts) {
+		const ExchangeMAPIStream *stream = (ExchangeMAPIStream *) body_parts->data;
+		part = camel_mime_part_new ();
+		if ((stream->proptag == PR_HTML || stream->proptag == PR_BODY_HTML_UNICODE)
+		    && inline_attachs) {
+			CamelMultipart *m_related;
+			m_related = mapi_mime_build_multipart_related (item, stream,
+								       inline_attachs);
+			camel_medium_set_content_object (CAMEL_MEDIUM (part),
+							 CAMEL_DATA_WRAPPER (m_related));
+			camel_object_unref (m_related);
+		} else
+			part = mapi_mime_msg_body (item, stream);
+		
+		camel_multipart_add_part (m_alternative, part);
+		camel_object_unref (part);
+	}
+
+	return m_alternative;
+}
+
+static CamelMultipart *
+mapi_mime_build_multipart_mixed (CamelMultipart *content, GSList *attachs)
+{
+	CamelMimePart *part = camel_mime_part_new ();
+	CamelMultipart *m_mixed = camel_multipart_new ();
+	camel_data_wrapper_set_mime_type (CAMEL_DATA_WRAPPER (m_mixed), 
+					  "multipart/mixed");
+	camel_multipart_set_boundary (m_mixed, NULL);
+
+	camel_medium_set_content_object (CAMEL_MEDIUM (part), CAMEL_DATA_WRAPPER (content));
+	camel_multipart_add_part (m_mixed, part);
+
+	if (attachs)
+		mapi_mime_multipart_add_attachments (m_mixed, attachs);
+
+	return m_mixed;
+}
+
+/*Takes raw attachment streams and converts to MIME Parts. Parts are added to
+  either inline / non-inline lists.*/
+static void
+mapi_mime_classify_attachments (GSList *attachments, GSList **inline_attachs, GSList **noninline)
+{
+	for (;attachments != NULL; attachments = attachments->next) {
+		ExchangeMAPIAttachment *attach = (ExchangeMAPIAttachment *)attachments->data;
+		ExchangeMAPIStream *stream = NULL;
+		const char *filename, *mime_type, *content_id = NULL; 
+		CamelContentType *content_type;
+		CamelMimePart *part;
+
+		stream = exchange_mapi_util_find_stream (attach->streams, PR_ATTACH_DATA_BIN);
+
+		if (!stream || stream->value->len <= 0) {
+			continue;
+		}
+
+		part = camel_mime_part_new ();
+
+		filename = (const char *) exchange_mapi_util_find_SPropVal_array_propval(attach->lpProps, 
+											 PR_ATTACH_LONG_FILENAME);
+
+		if (!(filename && *filename))
+			filename = (const char *) exchange_mapi_util_find_SPropVal_array_propval(attach->lpProps, 
+												 PR_ATTACH_FILENAME);
+		camel_mime_part_set_filename(part, g_strdup(filename));
+		camel_content_type_set_param (((CamelDataWrapper *) part)->mime_type, "name", filename);
+
+		/*Content-Type*/
+		mime_type = (const char *) exchange_mapi_util_find_SPropVal_array_propval (attach->lpProps, PR_ATTACH_MIME_TAG);
+		if (!mime_type)
+			mime_type = "application/octet-stream";
+
+		camel_mime_part_set_content (part, (const char *) stream->value->data, stream->value->len, mime_type);
+
+
+		content_type = camel_mime_part_get_content_type (part);
+		if (content_type && camel_content_type_is (content_type, "text", "*"))
+			camel_mime_part_set_encoding (part, CAMEL_TRANSFER_ENCODING_QUOTEDPRINTABLE);
+		else
+			camel_mime_part_set_encoding (part, CAMEL_TRANSFER_ENCODING_BASE64);
+
+		/*Content-ID*/
+		content_id = (const char *) exchange_mapi_util_find_SPropVal_array_propval(attach->lpProps, 
+											   PR_ATTACH_CONTENT_ID);
+		/* TODO : Add disposition */
+		if (content_id) {
+			camel_mime_part_set_content_id (part, content_id);
+			*inline_attachs = g_slist_append (*inline_attachs, part);
+		} else
+			*noninline = g_slist_append (*noninline, part);
+	}
+}
+
 static CamelMimeMessage *
-mapi_folder_item_to_msg( CamelFolder *folder,
-			 MapiItem *item,
-			 CamelException *ex )
+mapi_folder_item_to_msg( CamelFolder *folder, MapiItem *item, CamelException *ex )
 {
 	CamelMimeMessage *msg = NULL;
-	CamelMultipart *multipart = NULL;
-	CamelMimePart *part = NULL;
-	CamelDataWrapper *msg_content = NULL;
+	CamelMultipart *multipart_body = NULL;
 
 	GSList *attach_list = NULL;
-	GSList *body_part_list = NULL;
+	GSList *inline_attachs =  NULL; /*Used for mulitpart/related*/
+	GSList *noninline_attachs = NULL; 
+
+	gboolean build_alternative = FALSE;
+	gboolean build_related = FALSE;
 
 	attach_list = item->attachments;
 	msg = camel_mime_message_new ();
 
-	/* Threading */
-	if (item->header.message_id)
-		camel_medium_add_header (CAMEL_MEDIUM (msg), "Message-ID", item->header.message_id);
+	mapi_mime_set_recipient_list (msg, item);
+	mapi_mime_set_msg_headers (folder, msg, item);
+	mapi_mime_classify_attachments (attach_list, &inline_attachs, &noninline_attachs);
 
-	if (item->header.references)
-		camel_medium_add_header (CAMEL_MEDIUM (msg), "References", item->header.references);
+	build_alternative = (g_slist_length (item->msg.body_parts) > 1) && inline_attachs;
+	build_related = !build_alternative && inline_attachs;
 
-	if (item->header.in_reply_to)
-		camel_medium_add_header (CAMEL_MEDIUM (msg), "In-Reply-To", item->header.in_reply_to);
-
-	/* Set recipient details */
-	mapi_msg_set_recipient_list (msg, item);
-	mapi_populate_details_from_item (folder, msg, item);
-
-	if (g_slist_length (attach_list) > 0) {
-		multipart = camel_multipart_new ();
-		camel_data_wrapper_set_mime_type (CAMEL_DATA_WRAPPER (multipart), "multipart/mixed");
-		camel_multipart_set_boundary (multipart, NULL);
-
-		msg_content = CAMEL_DATA_WRAPPER (multipart);
+	if (build_alternative) {
+		multipart_body = mapi_mime_build_multipart_alternative (item, item->msg.body_parts,
+									inline_attachs);
+	} else if (build_related) {
+		multipart_body = mapi_mime_build_multipart_related (item,
+								    item->msg.body_parts->data,
+								    inline_attachs);
+	} else { /* Simple multipart/mixed */
+		CamelMimePart *part;
+		multipart_body = camel_multipart_new ();
+		camel_data_wrapper_set_mime_type (CAMEL_DATA_WRAPPER (multipart_body),
+						  "multipart/mixed");
+		camel_multipart_set_boundary (multipart_body, NULL);
+		part = mapi_mime_msg_body (item, item->msg.body_parts ? item->msg.body_parts->data : NULL);
+		camel_multipart_add_part (multipart_body, part);
+		camel_object_unref (part);
 	}
 
-	if (g_slist_length (item->msg.body_parts) > 1) {
-		multipart = camel_multipart_new ();
-		camel_data_wrapper_set_mime_type (CAMEL_DATA_WRAPPER (multipart), "multipart/alternative");
-		camel_multipart_set_boundary (multipart, NULL);
-
-		item->msg.body_parts = g_slist_sort (item->msg.body_parts, sort_bodies_cb);
-
-		/* Handle Multipart */
-		for (body_part_list = item->msg.body_parts; body_part_list; body_part_list = g_slist_next (body_part_list)) {
-			part = camel_mime_part_new ();
-
-			mapi_populate_msg_body_from_item (part, item, body_part_list->data);
-
-			camel_multipart_add_part (CAMEL_MULTIPART (msg_content), part);
-			camel_object_unref (part);
-		}
-
-		if (msg_content) {
-			part = camel_mime_part_new ();
-			camel_data_wrapper_set_mime_type (CAMEL_DATA_WRAPPER (part), "multipart/alternative");
-			camel_medium_set_content_object (CAMEL_MEDIUM (part), CAMEL_DATA_WRAPPER (multipart));
-			camel_object_unref (multipart);
-
-			camel_multipart_add_part (CAMEL_MULTIPART (msg_content), part);
-			camel_object_unref (part);
-		} else {
-			msg_content = CAMEL_DATA_WRAPPER (multipart);
-		}
-	} else {
-		if (msg_content) {
-			part = camel_mime_part_new ();
-			mapi_populate_msg_body_from_item (part, item, item->msg.body_parts ? item->msg.body_parts->data : NULL);
-			camel_multipart_add_part (CAMEL_MULTIPART (msg_content), part);
-			camel_object_unref (part);
-		} else {
-			msg_content = CAMEL_DATA_WRAPPER (msg);
-			mapi_populate_msg_body_from_item (CAMEL_MIME_PART (msg), item, item->msg.body_parts ? item->msg.body_parts->data : NULL);
-		}
+	if (noninline_attachs) { /* multipart/mixed */
+		multipart_body = mapi_mime_build_multipart_mixed (multipart_body, 
+								  noninline_attachs);
 	}
 
-	/* Attachment Handling */
-	if (attach_list) {
-		GSList *al;
-
-		multipart = CAMEL_MULTIPART (msg_content);
-
-		for (al = attach_list; al != NULL; al = al->next) {
-			ExchangeMAPIAttachment *attach = (ExchangeMAPIAttachment *)al->data;
-			ExchangeMAPIStream *stream = NULL;
-			const char *filename, *mime_type, *content_id = NULL; 
-			CamelContentType *content_type;
-
-			stream = exchange_mapi_util_find_stream (attach->streams, PR_ATTACH_DATA_BIN);
-
-			if (!stream || stream->value->len <= 0) {
-				continue;
-			}
-
-			part = camel_mime_part_new ();
-
-			filename = (const char *) exchange_mapi_util_find_SPropVal_array_propval(attach->lpProps, 
-												 PR_ATTACH_LONG_FILENAME);
-
-			if (!(filename && *filename))
-				filename = (const char *) exchange_mapi_util_find_SPropVal_array_propval(attach->lpProps, 
-													 PR_ATTACH_FILENAME);
-			camel_mime_part_set_filename(part, g_strdup(filename));
-			camel_content_type_set_param (((CamelDataWrapper *) part)->mime_type, "name", filename);
-
-			/*Content-Type*/
-			mime_type = (const char *) exchange_mapi_util_find_SPropVal_array_propval (attach->lpProps, PR_ATTACH_MIME_TAG);
-			if (!mime_type)
-				mime_type = "application/octet-stream";
-
-			camel_mime_part_set_content (part, (const char *) stream->value->data, stream->value->len, mime_type);
-
-			/*Content-ID*/
-			content_id = (const char *) exchange_mapi_util_find_SPropVal_array_propval(attach->lpProps, 
-												   PR_ATTACH_CONTENT_ID);
-
-			camel_mime_part_set_content_id (part, content_id);
-
-			content_type = camel_mime_part_get_content_type (part);
-			if (content_type && camel_content_type_is (content_type, "text", "*"))
-				camel_mime_part_set_encoding (part, CAMEL_TRANSFER_ENCODING_QUOTEDPRINTABLE);
-			else
-				camel_mime_part_set_encoding (part, CAMEL_TRANSFER_ENCODING_BASE64);
-
-			camel_multipart_add_part (multipart, part);
-			camel_object_unref (part);
-			
-		}
-		exchange_mapi_util_free_attachment_list (&attach_list);
-	}
-
-	if (msg_content != CAMEL_DATA_WRAPPER (msg)) {
-		camel_medium_set_content_object (CAMEL_MEDIUM (msg), msg_content);
-		camel_object_unref (msg_content);
-	}
+	camel_medium_set_content_object(CAMEL_MEDIUM (msg), CAMEL_DATA_WRAPPER(multipart_body));
+	camel_object_unref (multipart_body);
 
 	return msg;
 }
@@ -1768,11 +1882,13 @@ mapi_folder_get_message( CamelFolder *folder, const char *uid, CamelException *e
 		options |= MAPI_OPTIONS_USE_PFSTORE;
 	} 
 
+	CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
 	exchange_mapi_connection_fetch_item (id_folder, id_message, 
 					camel_GetPropsList, G_N_ELEMENTS (camel_GetPropsList), 
 					camel_build_name_id, NULL, 
 					fetch_item_cb, &item, 
 					options);
+	CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
 
 	if (item == NULL) {
 		camel_exception_set (ex, CAMEL_EXCEPTION_SERVICE_INVALID, _("Could not get message"));
@@ -1781,10 +1897,7 @@ mapi_folder_get_message( CamelFolder *folder, const char *uid, CamelException *e
 	}
 
 	msg = mapi_folder_item_to_msg (folder, item, ex);
-
-	exchange_mapi_util_free_recipient_list (&item->recipients);
-
-	g_free (item);
+	mapi_item_free (item);
 
 	if (!msg) {
 		camel_exception_set (ex, CAMEL_EXCEPTION_SERVICE_INVALID, _("Could not get message"));
@@ -1851,6 +1964,8 @@ mapi_get_message_info(CamelFolder *folder, const char *uid)
 		mi = (CamelMessageInfoBase *)msg_info ;
 		return (msg_info);
 	}
+	/* Go online and fetch message summary. */
+
 	msg_info = camel_message_info_new(folder->summary);
 	mi = (CamelMessageInfoBase *)msg_info ;
 
@@ -2053,19 +2168,23 @@ mapi_append_message (CamelFolder *folder, CamelMimeMessage *message,
 {
 	CamelMapiStore *mapi_store= CAMEL_MAPI_STORE(folder->parent_store);
 	CamelOfflineStore *offline = (CamelOfflineStore *) folder->parent_store;
-	CamelAddress *from = NULL, *recipients = NULL;
+	CamelAddress *from = NULL;
 	CamelStoreInfo *si;
 
 	MapiItem *item = NULL;
 	mapi_id_t fid = 0, mid = 0;
 	const gchar *folder_id;
+	guint32 folder_flags = 0;
 
 	/*Reject outbox / sent & trash*/
-	si = camel_store_summary_path ((CamelStoreSummary *)mapi_store->summary,
-				       folder->full_name);
+	si = camel_store_summary_path ((CamelStoreSummary *)mapi_store->summary, folder->full_name);
+	if (si) {
+		folder_flags = si->flags;
+		camel_store_summary_info_free ((CamelStoreSummary *)mapi_store->summary, si);
+	}
 
-	if (((si->flags & CAMEL_FOLDER_TYPE_MASK) == CAMEL_FOLDER_TYPE_TRASH) ||
-	    ((si->flags & CAMEL_FOLDER_TYPE_MASK) == CAMEL_FOLDER_TYPE_OUTBOX)) {
+	if (((folder_flags & CAMEL_FOLDER_TYPE_MASK) == CAMEL_FOLDER_TYPE_TRASH) ||
+	    ((folder_flags & CAMEL_FOLDER_TYPE_MASK) == CAMEL_FOLDER_TYPE_OUTBOX)) {
 		camel_exception_setv (ex, CAMEL_EXCEPTION_SYSTEM, 
 				      _("Cannot append message to folder '%s'"),
 				      folder->full_name);
@@ -2074,7 +2193,7 @@ mapi_append_message (CamelFolder *folder, CamelMimeMessage *message,
 
 	if (offline->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL) {
 		camel_exception_setv (ex, CAMEL_EXCEPTION_SYSTEM, 
-				 _("Offline. '%s'"));
+				      _("Offline."));
 		return;
 	}
 
@@ -2085,9 +2204,8 @@ mapi_append_message (CamelFolder *folder, CamelMimeMessage *message,
 
 	/* Convert MIME to Item */
 	from = (CamelAddress *) camel_mime_message_get_from (message);
-	recipients = (CamelAddress *) camel_mime_message_get_recipients (message, "to");
 
-	item = camel_mapi_utils_mime_to_item (message, from, recipients, ex);
+	item = camel_mapi_utils_mime_to_item (message, from, ex);
 
 	mid = exchange_mapi_create_item (-1, fid, NULL, NULL, 
 					 camel_mapi_utils_create_item_build_props, item,
@@ -2112,7 +2230,7 @@ camel_mapi_folder_class_init (CamelMapiFolderClass *camel_mapi_folder_class)
 	camel_folder_class->search_by_expression = mapi_folder_search_by_expression;
 	camel_folder_class->cmp_uids = mapi_cmp_uids;
 /* 	camel_folder_class->get_message_info = mapi_get_message_info; */
-/* 	camel_folder_class->search_by_uids = mapi_folder_search_by_uids;  */
+	camel_folder_class->search_by_uids = mapi_folder_search_by_uids;
 	camel_folder_class->search_free = mapi_folder_search_free;
 	camel_folder_class->append_message = mapi_append_message;
 	camel_folder_class->refresh_info = mapi_refresh_info;
@@ -2244,5 +2362,3 @@ camel_mapi_folder_new (CamelStore *store, const char *folder_name, const char *f
 	}
 	return folder;
 }
-
-
