@@ -77,12 +77,59 @@ void mapi_debug_logger_muted (const gchar* domain, GLogLevelFlags level,
 	/*Nothing here. Just a dummy function*/
 }
 
+static gboolean
+ensure_mapi_init_called (void)
+{
+	static gboolean called = FALSE;
+	gchar *profpath;
+	enum MAPISTATUS status;
+
+	LOCK ();
+	if (called) {
+		UNLOCK ();
+		return TRUE;
+	}
+
+	profpath = g_build_filename (g_get_home_dir (), DEFAULT_PROF_PATH, NULL);
+
+	if (!g_file_test (profpath, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR)) {
+		/* Create a ProfileStore */
+		status = CreateProfileStore (profpath, LIBMAPI_LDIF_DIR);
+		if (status != MAPI_E_SUCCESS && (status != MAPI_E_NO_ACCESS || !g_file_test (profpath, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR))) {
+			mapi_errstr ("CreateProfileStore", GetLastError());
+			g_free (profpath);
+
+			UNLOCK ();
+			return FALSE;
+		}
+	}
+
+	status = MAPIInitialize (profpath);
+	if (status == MAPI_E_SESSION_LIMIT) {
+		/* do nothing, the profile store is already initialized */
+		/* but this shouldn't happen */
+		mapi_errstr ("MAPIInitialize", GetLastError());
+	} else if (status != MAPI_E_SUCCESS) {
+		mapi_errstr ("MAPIInitialize", GetLastError());
+		g_free (profpath);
+
+		UNLOCK ();
+		return FALSE;
+	}
+
+	g_free (profpath);
+
+	called = TRUE;
+	UNLOCK ();
+
+	return TRUE;
+}
+
 static struct mapi_session *
 mapi_profile_load (const char *profname, const char *password)
 {
 	enum MAPISTATUS	retval = MAPI_E_SUCCESS;
 	struct mapi_session *session = NULL;
-	gchar *profpath = NULL;
 	gchar *default_profile_name = NULL;
 	const char *profile = NULL;
 	guint32 debug_log_level = 0;
@@ -95,24 +142,11 @@ mapi_profile_load (const char *profname, const char *password)
 
 	g_debug("%s: Entering %s ", G_STRLOC, G_STRFUNC);
 
-	profpath = g_build_filename (g_get_home_dir(), DEFAULT_PROF_PATH, NULL);
-	if (!g_file_test (profpath, G_FILE_TEST_EXISTS)) {
-		g_debug ("MAPI profile database @ %s not found ", profpath);
+	if (!ensure_mapi_init_called ())
 		goto cleanup;
-	}
-
-	MAPIUninitialize ();
-
-	retval = MAPIInitialize(profpath);
-	if (retval != MAPI_E_SUCCESS) {
-		mapi_errstr("MAPIInitialize", GetLastError());
-		if (retval == MAPI_E_SESSION_LIMIT)
-			g_debug("%s: %s: Already connected ", G_STRLOC, G_STRFUNC);
-		goto cleanup;
-	}
 
 	/* Initialize libmapi logger*/
-	if (g_getenv ("LIBMAPI_DEBUG")) {
+	if (g_getenv ("MAPI_DEBUG")) {
 		debug_log_level = atoi (g_getenv ("MAPI_DEBUG"));
 		SetMAPIDumpData(TRUE);
 		SetMAPIDebugLevel(debug_log_level);
@@ -138,12 +172,6 @@ mapi_profile_load (const char *profname, const char *password)
 	}
 
 cleanup:
-	if (retval != MAPI_E_SUCCESS && retval != MAPI_E_SESSION_LIMIT &&
-	    retval != MAPI_E_LOGON_FAILED)
-		MAPIUninitialize ();
-
-	g_free (profpath);
-
 	g_debug("%s: Leaving %s ", G_STRLOC, G_STRFUNC);
 
 	return session;
@@ -172,13 +200,27 @@ exchange_mapi_connection_new (const char *profile, const char *password)
 }
 
 void
-exchange_mapi_connection_close ()
+exchange_mapi_connection_close (void)
 {
 	LOCK();
+
+	if (global_mapi_session) {
+		mapi_object_t obj_store;
+		enum MAPISTATUS status;
+
+		mapi_object_init (&obj_store);
+
+		/* Open the message store */
+		status = OpenMsgStore (global_mapi_session, &obj_store);
+		if (status != MAPI_E_SUCCESS) {
+			mapi_errstr ("OpenMsgStore", GetLastError());
+		} else {
+			Logoff (&obj_store);
+		}
+	}
 	global_mapi_session = NULL;
-	MAPIUninitialize ();	
+
 	UNLOCK();
-	/* TODO :  Return status. get last error ? */
 }
 
 static gboolean 
@@ -252,7 +294,7 @@ exchange_mapi_util_read_generic_stream (mapi_object_t *obj_message, uint32_t pro
 
 		/* Build a mapi_SPropValue_array structure */
 		properties_array.cValues = 1; 
-		properties_array.lpProps = talloc_array (mem_ctx, struct mapi_SPropValue, properties_array.cValues);
+		properties_array.lpProps = talloc_zero_array (mem_ctx, struct mapi_SPropValue, properties_array.cValues + 1);
 		properties_array.lpProps[0].ulPropTag = proptag; 
 		/* This call is needed in case the read stream was a named prop. */
 		mapi_SPropValue_array_named (obj_message, &properties_array);
@@ -857,8 +899,9 @@ exchange_mapi_util_get_recipients (mapi_object_t *obj_message, GSList **recip_li
 			mapidump_SRow (&(rows_recip.aRow[i_row_recip]), " ");
 		}
 
-		recipient->out.all_lpProps = talloc_steal ((TALLOC_CTX *)recipient->mem_ctx, rows_recip.aRow[i_row_recip].lpProps);
-		recipient->out.all_cValues = rows_recip.aRow[i_row_recip].cValues;
+		recipient->out_SRow.ulAdrEntryPad = rows_recip.aRow[i_row_recip].ulAdrEntryPad;
+		recipient->out_SRow.cValues = rows_recip.aRow[i_row_recip].cValues;
+		recipient->out_SRow.lpProps = talloc_steal ((TALLOC_CTX *)recipient->mem_ctx, rows_recip.aRow[i_row_recip].lpProps);
 
 		*recip_list = g_slist_append (*recip_list, recipient);
 	}
@@ -1118,7 +1161,7 @@ exchange_mapi_connection_fetch_items   (mapi_id_t fid,
 	mapi_object_t obj_store;
 	mapi_object_t obj_folder;
 	mapi_object_t obj_table;
-	struct SPropTagArray *SPropTagArray, *GetPropsTagArray;
+	struct SPropTagArray *SPropTagArray, *GetPropsTagArray = NULL;
 	struct SRowSet SRowSet;
 	uint32_t count, i, cursor_pos = 0;
 	gboolean result = FALSE;
@@ -1154,9 +1197,6 @@ exchange_mapi_connection_fetch_items   (mapi_id_t fid,
 		goto cleanup;
 	}
 
-	GetPropsTagArray = talloc_zero(mem_ctx, struct SPropTagArray);
-	GetPropsTagArray->cValues = 0;
-
 	SPropTagArray = set_SPropTagArray(mem_ctx, 0x4,
 					  PR_FID,
 					  PR_MID,
@@ -1187,6 +1227,53 @@ exchange_mapi_connection_fetch_items   (mapi_id_t fid,
 		}
 	}
 
+	if ((GetPropsList && (cn_props > 0)) || build_name_id) {
+		struct SPropTagArray *NamedPropsTagArray;
+		uint32_t m;
+		struct mapi_nameid *nameid;
+
+		nameid = mapi_nameid_new(mem_ctx);
+		NamedPropsTagArray = talloc_zero(mem_ctx, struct SPropTagArray);
+
+		NamedPropsTagArray->cValues = 0;
+		/* Add named props using callback */
+		if (build_name_id) {
+			if (!build_name_id (nameid, build_name_data)) {
+				g_debug ("%s: (%s): Could not build named props ",
+					 G_STRLOC, G_STRFUNC);
+				goto GetProps_cleanup;
+			}
+
+			retval = mapi_nameid_GetIDsFromNames(nameid, &obj_folder, NamedPropsTagArray);
+			if (retval != MAPI_E_SUCCESS) {
+				mapi_errstr("mapi_nameid_GetIDsFromNames", GetLastError());
+				goto GetProps_cleanup;
+			}
+		}
+
+GetProps_cleanup:
+		for (m = 0; m < NamedPropsTagArray->cValues; m++) {
+			if (G_UNLIKELY (!GetPropsTagArray))
+				GetPropsTagArray = set_SPropTagArray (mem_ctx, 0x1,
+								      NamedPropsTagArray->aulPropTag[m]);
+			else
+				SPropTagArray_add (mem_ctx, GetPropsTagArray,
+						   NamedPropsTagArray->aulPropTag[m]);
+		}
+
+		for (m = 0; m < cn_props; m++) {
+			if (G_UNLIKELY (!GetPropsTagArray))
+				GetPropsTagArray = set_SPropTagArray (mem_ctx, 0x1,
+								      GetPropsList[m]);
+			else
+				SPropTagArray_add (mem_ctx, GetPropsTagArray,
+						   GetPropsList[m]);
+		}
+
+		MAPIFreeBuffer (NamedPropsTagArray);
+		talloc_free (nameid);
+	}
+
 	/* Note : We maintain a cursor position. count parameter in QueryRows */
 	/* is more of a request and not gauranteed  */
 	do {
@@ -1204,44 +1291,6 @@ exchange_mapi_connection_fetch_items   (mapi_id_t fid,
 			goto cleanup;
 		}
 
-		if ((GetPropsList && (cn_props > 0)) || build_name_id) {
-			struct SPropTagArray *NamedPropsTagArray;
-			uint32_t m, n=0;
-			struct mapi_nameid *nameid;
-
-			nameid = mapi_nameid_new(mem_ctx);
-			NamedPropsTagArray = talloc_zero(mem_ctx, struct SPropTagArray);
-
-			NamedPropsTagArray->cValues = 0;
-			/* Add named props using callback */
-			if (build_name_id) {
-				if (!build_name_id (nameid, build_name_data)) {
-					g_debug ("%s: (%s): Could not build named props ",
-							G_STRLOC, G_STRFUNC);
-					goto GetProps_cleanup;
-				}
-
-				retval = mapi_nameid_GetIDsFromNames(nameid, &obj_folder, NamedPropsTagArray);
-				if (retval != MAPI_E_SUCCESS) {
-					mapi_errstr("mapi_nameid_GetIDsFromNames", GetLastError());
-					goto GetProps_cleanup;
-				}
-			}
-
-			GetPropsTagArray->cValues = (cn_props + NamedPropsTagArray->cValues);
-			GetPropsTagArray->aulPropTag = talloc_array(mem_ctx, uint32_t, (cn_props + NamedPropsTagArray->cValues));
-
-			for (m = 0; m < NamedPropsTagArray->cValues; m++, n++)
-				GetPropsTagArray->aulPropTag[n] = NamedPropsTagArray->aulPropTag[m];
-
-			for (m = 0; m < cn_props; m++, n++)
-				GetPropsTagArray->aulPropTag[n] = GetPropsList[m];
-
-		GetProps_cleanup:
-			MAPIFreeBuffer (NamedPropsTagArray);
-			talloc_free (nameid);
-		}
-
 		for (i = 0; i < SRowSet.cRows; i++) {
 			mapi_object_t obj_message;
 			struct mapi_SPropValue_array properties_array = {0};
@@ -1283,16 +1332,21 @@ exchange_mapi_connection_fetch_items   (mapi_id_t fid,
 				exchange_mapi_util_read_body_stream (&obj_message, &stream_list, 
 								     options & MAPI_OPTIONS_GETBESTBODY);
 
-			if (GetPropsTagArray->cValues) {
+			if (GetPropsTagArray && GetPropsTagArray->cValues) {
 				struct SPropValue *lpProps;
+				struct SPropTagArray *tags;
 				uint32_t prop_count = 0, k;
-
-				retval = GetProps (&obj_message, GetPropsTagArray, &lpProps, &prop_count);
-
-				/* Conversion from SPropValue to mapi_SPropValue. (no padding here) */
+				/* we need to make a local copy of the tag array
+				 * since GetProps will modify the array on any
+				 * errors */
+				tags = set_SPropTagArray (mem_ctx, 0x1, GetPropsTagArray->aulPropTag[0]);
+				for (k = 1; k < GetPropsTagArray->cValues; k++)
+					SPropTagArray_add (mem_ctx, tags, GetPropsTagArray->aulPropTag[k]);
+				retval = GetProps (&obj_message, tags, &lpProps, &prop_count);
+				MAPIFreeBuffer (tags);
+				properties_array.lpProps = talloc_zero_array (mem_ctx, struct mapi_SPropValue, 
+									 prop_count + 1);
 				properties_array.cValues = prop_count;
-				properties_array.lpProps = talloc_array (mem_ctx, struct mapi_SPropValue, 
-									 prop_count);
 				for (k=0; k < prop_count; k++)
 					cast_mapi_SPropValue(&properties_array.lpProps[k], &lpProps[k]);
 
@@ -1335,7 +1389,7 @@ exchange_mapi_connection_fetch_items   (mapi_id_t fid,
 				exchange_mapi_util_free_attachment_list (&attach_list);
 			}
 
-			if (GetPropsTagArray->cValues) 
+			if (GetPropsTagArray && GetPropsTagArray->cValues) 
 				talloc_free (properties_array.lpProps);
 
 		loop_cleanup:
@@ -1350,6 +1404,8 @@ exchange_mapi_connection_fetch_items   (mapi_id_t fid,
 	result = TRUE;
 
 cleanup:
+	if (GetPropsTagArray)
+		MAPIFreeBuffer (GetPropsTagArray);
 	mapi_object_release(&obj_folder);
 	mapi_object_release(&obj_table);
 	mapi_object_release(&obj_store);
@@ -1475,7 +1531,7 @@ exchange_mapi_connection_fetch_item (mapi_id_t fid, mapi_id_t mid,
 
 		/* Conversion from SPropValue to mapi_SPropValue. (no padding here) */
 		properties_array.cValues = prop_count;
-		properties_array.lpProps = talloc_array (mem_ctx, struct mapi_SPropValue, prop_count);
+		properties_array.lpProps = talloc_zero_array (mem_ctx, struct mapi_SPropValue, prop_count + 1);
 		for (k=0; k < prop_count; k++)
 			cast_mapi_SPropValue(&properties_array.lpProps[k], &lpProps[k]);
 
@@ -3120,11 +3176,14 @@ exchange_mapi_events_subscribe (guint32 options,
 
 	mapi_object_init(&obj_target);
 
+	LOCK ();
+
 	if (options & MAPI_EVENTS_USE_STORE) {
 		retval = OpenMsgStore(global_mapi_session, &obj_target);
 		if (retval != MAPI_E_SUCCESS) {
 			mapi_errstr("OpenMsgStore", GetLastError());
-			return false;
+			UNLOCK ();
+			return FALSE;
 		}
 	} else if (options & MAPI_EVENTS_USE_PF_STORE) {
 		/* TODO */
@@ -3135,6 +3194,8 @@ exchange_mapi_events_subscribe (guint32 options,
 	retval = Subscribe(&obj_target, connection, event_mask, use_store,
 			   (mapi_notify_callback_t) callback, data);
 
+	UNLOCK ();
+
 	return (retval == MAPI_E_SUCCESS);
 }
 
@@ -3143,11 +3204,16 @@ exchange_mapi_events_unsubscribe (guint32 connection)
 {
 	enum MAPISTATUS	retval;
 
-	if (!global_mapi_session)
+	LOCK ();
+	if (!global_mapi_session) {
+		UNLOCK ();
 		return FALSE;
+	}
 
 	retval = Unsubscribe (global_mapi_session, connection);
 
+	UNLOCK ();
+
 	return (retval == MAPI_E_SUCCESS);
 }
 
@@ -3187,7 +3253,7 @@ exchange_mapi_create_profile (const char *username, const char *password, const
 	enum MAPISTATUS	retval;
 	gboolean result = FALSE; 
 	const gchar *workstation = "localhost";
-	gchar *profname = NULL, *profpath = NULL;
+	gchar *profname = NULL;
 	struct mapi_session *session = NULL;
 
 	/*We need all the params before proceeding.*/
@@ -3198,31 +3264,9 @@ exchange_mapi_create_profile (const char *username, const char *password, const
 
 	LOCK ();
 
-	profpath = g_build_filename (g_get_home_dir(), DEFAULT_PROF_PATH, NULL);
 	profname = g_strdup_printf("%s@%s", username, domain);
 
-	if (!g_file_test (profpath, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR)) {
-		/* Create a ProfileStore */
-		retval = CreateProfileStore (profpath, LIBMAPI_LDIF_DIR); 
-		if (retval != MAPI_E_SUCCESS) {
-			manage_mapi_error ("CreateProfileStore", GetLastError(), error_msg);
-			g_free (profpath);
-			g_free (profname);
-			
-			UNLOCK ();
-			return FALSE;
-		}
-	}
-
-	retval = MAPIInitialize(profpath); 
-	if (retval == MAPI_E_SESSION_LIMIT)
-	/* do nothing, the profile store is already initialized */
-		manage_mapi_error ("MAPIInitialize", GetLastError(), error_msg); 
-	else if (retval != MAPI_E_SUCCESS) {
-		manage_mapi_error ("MAPIInitialize", GetLastError(), error_msg);
-		g_free (profpath);
-		g_free (profname);
-		
+	if (!ensure_mapi_init_called ()) {
 		UNLOCK ();
 		return FALSE;
 	}
@@ -3285,13 +3329,9 @@ exchange_mapi_create_profile (const char *username, const char *password, const
 	} else 
 		goto exit;
 
-cleanup: 
-	if (!result)
-		MAPIUninitialize ();
-
-exit:
+ cleanup: 
+ exit:
 	g_free (profname);
-	g_free (profpath);
 
 	UNLOCK ();
 
@@ -3303,23 +3343,10 @@ exchange_mapi_delete_profile (const char *profile)
 {
 	enum MAPISTATUS	retval;
 	gboolean result = FALSE; 
-	gchar *profpath = NULL;
 
 	LOCK ();
 
-	profpath = g_build_filename (g_get_home_dir(), DEFAULT_PROF_PATH, NULL);
-	if (!g_file_test (profpath, G_FILE_TEST_EXISTS)) {
-		g_debug ("No need to delete profile. DB itself is missing \n");
-		result = TRUE;
-		goto cleanup; 
-	}
-
-	retval = MAPIInitialize(profpath); 
-	if (retval == MAPI_E_SESSION_LIMIT)
-	/* do nothing, the profile store is already initialized */
-		; 
-	else if (retval != MAPI_E_SUCCESS) {
-		mapi_errstr("MAPIInitialize", GetLastError());
+	if (!ensure_mapi_init_called ()) {
 		goto cleanup; 
 	}
 
@@ -3334,8 +3361,6 @@ exchange_mapi_delete_profile (const char *profile)
 	result = TRUE; 
 
 cleanup: 
-	g_free(profpath);
-
 	UNLOCK ();
 
 	return result;
